
from lang.cfg import CFG, InnerNode, Node
from lang.ast import AST, Assignment, Branch, Statement, WhileLoop, Guard

from pathlib import Path

from lark import Lark, ast_utils
from lang.ast import ToAST


def fileToCFG(filename: str) -> CFG:
    with open(filename) as file:
        content = "\n".join(file.readlines())
        return stringToCFG(content)


def stringToCFG(s: str) -> CFG:
    transformer = ast_utils.create_transformer("tundra", ToAST())
    parser = Lark.open("tundra.lark", rel_to=Path(
        __file__).absolute(), start='program', propagate_positions=True)
    ast = transformer.transform(parser.parse(s))
    return astToCFG(ast)


def astToCFG(ast: AST) -> CFG:
    cfg = CFG(ast.variables)
    last_nodes = addCFGEdgesForStatementList(ast.statements, cfg, {cfg.entry})
    for last_node in last_nodes:
        cfg.next[last_node].add(cfg.exit)
        cfg.previous[cfg.exit].add(last_node)
    return cfg


def addCFGEdgesForStatementList(lst: list[Statement], cfg: CFG, entries: set[Node]) -> set[Node]:
    """
    This function updates cfg to add nodes and edges corresponding to the statements in lst.
    Edges are also added from the nodes in entries to nodes corresponding to the statements in lst.

    Note that only Guard and Assignment statements are added to inner nodes. 
    Other statements (WhileLoop and Branch) are represented using Guards. 

    Return the set of nodes corresponding to the exit locations of this block.
    """

    for entry in entries:
        assert (entry in cfg.nodes)

    previous_nodes = entries

    for i, stmt in enumerate(lst):
        if isinstance(stmt, Assignment):
            new_node = InnerNode(stmt)
            cfg.add_node(new_node)
            for node in previous_nodes:
                cfg.next[node].add(new_node)
                cfg.previous[new_node].add(node)
            previous_nodes = {new_node}
        elif isinstance(stmt, WhileLoop):

            new_node = InnerNode(Guard(stmt.condition, stmt.label))
            cfg.add_node(new_node)
            # Add edges for the while loop's body
            # and connect them to the loop's condition node
            block_exits = addCFGEdgesForStatementList(
                stmt.statements, cfg, {new_node})
            # Add edges connecting the exits of the body back
            # to the condition node.
            for node in block_exits:
                cfg.next[node].add(new_node)
                cfg.previous[new_node].add(node)
            # Connect previous nodes to the condition node
            for node in previous_nodes:
                cfg.next[node].add(new_node)
                cfg.previous[new_node].add(node)
            previous_nodes = {new_node}
        elif isinstance(stmt, Branch):
            new_node = InnerNode(Guard(stmt.condition, stmt.label))
            cfg.add_node(new_node)
            br1_ex = addCFGEdgesForStatementList(
                stmt.branch1, cfg, {new_node})
            br2_ex = addCFGEdgesForStatementList(
                stmt.branch2, cfg, {new_node})
            for node in previous_nodes:
                cfg.next[node].add(new_node)
                cfg.previous[new_node].add(node)
            # The exit locations of each branch are combined
            # to form the exit locations of this branch statement.
            previous_nodes = br1_ex.union(br2_ex)

        else:
            raise Exception(
                f"Statement {stmt} not supported for addCFGEdgesForStatementList on statement list {lst}")

    return previous_nodes
