from analysis.fixed_point import fixed_point
from analysis.live_variables import LiveVariables
from lang.ast.ast import Variable
from lang.cfg import Exit
from lang.to_cfg import stringToCFG
from lang.utils import getEntryValue, getExitValue, getValueByLabel


def test_live_variables_sequential1():
    s = """
    x = 0; 
    y = 10;  
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    # There are no live variables at any program points 
    # because no variables are ever used. 
    assert(all([v == set() for v in In.values()]))
    assert(all([v == set() for v in Out.values()]))

def test_live_variables_sequential2():
    s = """
    l1: x = 0; 
    l2: y = x;  
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l1", Out) == { Variable("x")})
    assert(getValueByLabel("l2", Out) == set())

def test_live_variables_branch1():
    s = """
        y = 1; 
    l1: x = 0; 
    l2: if (x < 0) {
            x = 0; 
        } else {

        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l1", Out) == { Variable("x")})

def test_live_variables_branch2():
    s = """
        y = 1; 
    l1: x = 0; 
    l2: if (x < 0) {
            x = 0; 
        } else {
            y = 0;
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l1", Out) == { Variable("x")})

def test_live_variables_branch3():
    s = """
        y = 1; 
    l1: x = 0; 
    l2: if (x < 0) {
            x = 0; 
        } else {
            y = y + 1;
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l2", In) == { Variable("x"), Variable("y") })


def test_live_variables_branch3():
    s = """
        y = 1; 
    l1: x = 0; 
    l2: if (x < 0) {
            x = z; 
        } else {
            x = y;
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l1", Out) == { Variable("x"), Variable("y"), Variable("z")})

def test_live_variables_branch4():
    s = """
        y = 1; 
    l1: x = 0; 
    l2: if (x < 0) {
            x = z; 
        } else {
            x = y;
        }
    l3: y = x; 
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l3", In) == { Variable("x")})
    assert(getValueByLabel("l3", Out) == set())

def test_live_variables_loop1():
    s = """
    l1: while (y < 0) {
            x = x + 1; 
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l1", In) == { Variable("x"), Variable("y") })
    assert(getValueByLabel("l1", Out) == { Variable("x"), Variable("y") })

def test_live_variables_loop2():
    s = """
        while ((y + x) < 0) {
            l1: x = y + 1; 
            l2: y = x + 1;
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getValueByLabel("l1", In) == { Variable("y") })
    assert(getValueByLabel("l1", Out) == { Variable("x") })
    assert(getValueByLabel("l2", In) == { Variable("x") })
    assert(getValueByLabel("l2", Out) == { Variable("x"), Variable("y") })

def test_live_variables_loop3():
    s = """
        while ((y + x) < 0) {
            l1: x = y + 1; 
            l2: y = x + 1;
        }
        l3: y = x;
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(getEntryValue(Out) == { Variable("x"), Variable("y") })
    assert(getExitValue(In) == set() )
    assert(getValueByLabel("l3", In) == { Variable("x") })

def test_live_variables_loop4():
    s = """
        l1: while (x < 0) {
            x = 0; 
            if (y < 0) {
                y = z + 1; 
            } else { }
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, LiveVariables(cfg))
    assert(all([Variable("z") in In[k] for k in In if not isinstance(k, Exit)]))

